package bemused;

import protocol.*;
import java.util.Vector;

/*
 * Created August 2004
 */

/**
 * @author johan khne
 */

public class Browser {

	BemusedProtocol protocol;
	String[] dirs;
	String[] files;
	String activeDir;
	
	boolean updateVisible;
	boolean dirChanged;
	
	
	public Browser(BemusedProtocol prot) {
		protocol = prot;
		activeDir = "";
		
		int tempsize = 1;
		dirs = new String[0];
		files = new String[0];
		
		updateVisible = true;
		
		dirChanged = false;
	}
	
	public String activeDir() {
		int lastSlash = 0, newLastSlash = 0;

		while (newLastSlash > -1) {
			lastSlash = newLastSlash;
			newLastSlash = activeDir.indexOf("\\", lastSlash + 1);
		}
		return activeDir.substring(lastSlash, activeDir.length());
	}
	
	public String[] dirs() {
		return dirs;
	}
	
	public String[] files() {
		return files;
	}
	
	public void readDir() {
		protocol.readDir(activeDir);
	}
	
	public int fetchIndex(String jumpTo) {
		int len = jumpTo.length();
		jumpTo = jumpTo.toLowerCase();
		
		for (int n = 0; n < dirs.length; n++) {
			if (dirs[n].substring(0, len).toLowerCase().equals(jumpTo)) {
				return n;
			}
		}
		for (int n = 0; n < files.length; n++) {
			if (files[n].substring(0, len).toLowerCase().equals(jumpTo)) {
				return (n + dirs.length);
			}
		}
		return 0;
	}

	public void setDirInfo(byte[] headerBytes, String[] struc) {
		// For more information, see:
		// http://bemused.sourceforge.net/book/view/49
	
		byte headerByte = headerBytes[0];
		String fileName;
		boolean bit0, bit1, bit2, bit3, bit4, bit5, bit6, bit7, bit8;
		boolean type_dir, type_file, type_expandable_dir;
		Vector dirList = new Vector();
		Vector fileList = new Vector();
		
		for (int n = 0; n < struc.length; n++) 	{
			if (n == 0) {
				if (headerByte / 15 == 0) { // top 4 bits is 0x0 = root node
					activeDir = struc[0];
					continue;
				}
				else { // normal entry
					activeDir = "";
				}
			}

			headerByte = headerBytes[n];
			fileName = struc[n];
			
			// hardcoded thumbs.db filter
			if (fileName.toLowerCase().equals("thumbs.db")) continue;

			// decode headerByte
			// there has to be an easier way :/
			
			bit4 = (headerByte & 8) == 8 ? true : false;
			bit5 = (headerByte & 4) == 4 ? true : false;
			bit6 = (headerByte & 2) == 2 ? true : false;
			bit7 = (headerByte & 1) == 1 ? true : false; 
			
			type_dir = (!bit4 && !bit5 && !bit6 && bit7) ? true : false;
			type_file = (!bit4 && !bit5 && !bit6 && bit7) ? true : false;
			type_expandable_dir = (!bit4 && !bit5 && bit6 && bit7) ? true : false;
			
			if (type_dir || type_expandable_dir) {
				dirList.addElement(fileName);
			}
			else { // type_file
				fileList.addElement(fileName);
			}
		}
		
		int listSize = dirList.size();
		dirs = new String[listSize < 1 ? 0 : listSize];
		for (int n = 0; n < listSize; n++) {
			dirs[n] = (String) dirList.elementAt(n);
		}
		
		listSize = fileList.size();
		files = new String[listSize < 1 ? 0 : listSize];
		for (int n = 0; n < listSize; n++) {
			files[n] = (String) fileList.elementAt(n);
		}
		
		quicksort(dirs);
		quicksort(files);
		
		dirChanged = true;
		
		if (updateVisible) {
			protocol.notifyStatusUpdate();
		}
	}
	
	public boolean dirChanged() {
		if (dirChanged) {
			dirChanged = false;
			return true;
		}
		return dirChanged;
	}
	
	public void addToPlaylist(String f) {
		protocol.addToPlaylist(f);
	}
	
	public void enqueue(int index) {
		if (index == -1) return;
		if (index >= dirs.length) {
			addToPlaylist(activeDir.equals("") ? files[index - dirs.length] : activeDir + "\\" + files[index - dirs.length]);
			protocol.infoAlert("Adding", "File added to playlist!", 1500);
		}
		else {
			addToPlaylist(activeDir.equals("") ? dirs[index] : activeDir + "\\" + dirs[index]);
			protocol.infoAlert("Adding", "Dir added to playlist!", 2000);
		}
	}
	
	public void openItem(int index) {
		if (index == -1) return;
		if (index >= dirs.length) {
			protocol.playFile(activeDir.equals("") ? files[index - dirs.length] : activeDir + "\\" + files[index - dirs.length]);
		}
		else {
			changeDir(dirs[index]);
		}
	}
	
	public void changeDir(String dir) {
		activeDir = activeDir.equals("") ? dir : activeDir + "\\" + dir;
		readDir();
	}

	public void moveDirUp() {
		int lastSlash = 0, newLastSlash = 0;
		while (newLastSlash > -1) {
			lastSlash = newLastSlash;
			newLastSlash = activeDir.indexOf("\\", lastSlash + 1);
		}
		activeDir = activeDir.substring(0, lastSlash);
		if (activeDir.indexOf("\\", 0) == -1) activeDir = "";
		readDir();
	}
	
	private static void quicksort(String[] data) {
		quicksort(data, 0, data.length-1);
	}
	
	private static void quicksort(String[] data, int lb, int ub) {
		if (ub <= lb) return;
		int part = partition(data, lb, ub);
		quicksort(data, lb, part-1);
		quicksort(data, part+1, ub);
	}
	
	private static int partition(String[] data, int lb, int ub) {
		while (true) {
			while (lb < ub && data[lb].toLowerCase().compareTo(data[ub].toLowerCase()) < 0) ub--;
			if (lb < ub) swap(data, lb++, ub);
			else return lb;
			
			while (lb < ub && data[lb].toLowerCase().compareTo(data[ub].toLowerCase()) < 0) lb++;
			if (lb < ub) swap(data, lb, ub--);
			else return ub;
		}
	}
	
	private static void swap(String[] data, int x, int y) {
		String tmp = data[x];
		data[x] = data[y];
		data[y] = tmp;
	}
}
