package bemused;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
// import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.lang.Byte;

import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.AlertType;

import protocol.*;
import protocol.MusicPlayer;

/*
 * Created on Apr 20, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
/**
 * @author fred
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class BemusedProtocol implements MusicPlayer {
	String connectionURL;
	StreamConnection connection;
	DataInputStream iStream;
	DataOutputStream oStream;
	Display display;
	Thread readWriteThread;
	boolean reconnect;
	Integer runSignal;
	Vector statusUpdateListeners;
	ProtocolStatus status;
	Vector cmdQueue;
	MyCommandTarget commandTarget;
	Timer statusUpdateTimer;
	public Browser fileBrowser;
	
	public class PlaybackStatus {
		public PlaybackStatus() {
			playing = false;
			repeat = false;
			shuffle = false;
			startTime = new Date();
			songLengthSecs = 0;
			title = "<no title>";
		}
		public boolean playing;
		public boolean shuffle;
		public boolean repeat;
		public Date startTime;
		public long songLengthSecs;
		public String title;
		public String[] playlist;
		public int playlistPos;
		public int volume;
	}
	PlaybackStatus currentStatus;
	
	public BemusedProtocol(Display display) {
		status = new ProtocolStatus();
		statusUpdateTimer = new Timer();
		currentStatus = new PlaybackStatus();
		commandTarget = new MyCommandTarget();
		currentStatus.playlist = new String[0];
		this.display = display;
		runSignal = new Integer(0);
		statusUpdateListeners = new Vector();
		cmdQueue = new Vector();
		
		fileBrowser = new Browser(this);

		Runnable runnable = new Runnable() {
			public void run() {
				BemusedProtocol.this.run();	
			}
		};
		reconnect = true;
		readWriteThread = new Thread(runnable);
		readWriteThread.start();

		// Update the status information every few seconds
		// TODO: Optimize this. Only update when we can expect a change
		scheduleStatusUpdate();
	}
	
	public void infoAlert(String title, String msg, int timeout) {
		Alert alert = new Alert (title, msg, null, AlertType.INFO);
		alert.setTimeout (timeout);
		display.setCurrent(alert);
	}
	
	void scheduleStatusUpdate(long milliSec) {
		
		statusUpdateTimer.cancel();
		statusUpdateTimer = new Timer();

		statusUpdateTimer.schedule( 
				new TimerTask() { public void run() {
					// Cancel pending timers
					appendCommand(new Command("INF2"));
					appendCommand(new Command("GVOL"));
					
					// Re-schedule update after the current song has finished
					// or after 10 seconds, whatever is earlier
					PlaybackStatus status = status();
					long trackEndTimeout = Long.MAX_VALUE;
					if (status.playing) {
						trackEndTimeout = status.startTime.getTime() + 
								(status.songLengthSecs+1)*1000 - new Date().getTime();
						if (trackEndTimeout < 1000) trackEndTimeout = 1000;
					}
					long regularUpdateInterval = 10*1000;
					if (regularUpdateInterval < trackEndTimeout) {						
						scheduleStatusUpdate(regularUpdateInterval);
					}
					else {
						scheduleStatusUpdate(trackEndTimeout);
					}
				}}, milliSec);
	}
	
	void scheduleStatusUpdate() {		
		// Schedule update in one second
		scheduleStatusUpdate(1000);		
	}
	
	/*public void init() throws IOException {
		iStream = connection.openDataInputStream();
		oStream = connection.openDataOutputStream();
		Runnable runnable = new Runnable() {
			public void run() {
				BemusedProtocol.this.run();	
			}
		};
		readWriteThread = new Thread(runnable);
		readWriteThread.start();
	}*/
	
	private void run() {
		while (reconnect == true) {
			try {
				synchronized (runSignal) {
					runSignal.wait();					
				}
				openConnection();
				do {
					doNextCommand();
					synchronized (runSignal) {
						runSignal.wait();
					}
				} while (reconnect == true);
			} catch (IOException e) {
				closeConnection();
				continue;			
			} catch (InterruptedException e) {
				e.printStackTrace();
				throw new RuntimeException("InterruptedExcpetion caught: "+e.getMessage());	
			}
		}
	}
	
	void openConnection() throws IOException {
		closeConnection();
		if (connectionURL != null) {
			connection = (StreamConnection)Connector.open(connectionURL);
			iStream = connection.openDataInputStream();
			oStream = connection.openDataOutputStream();
			status.connected = true;
			notifyStatusUpdate();
		}
		else {
			throw new IOException("Bemused connection URL is null");
		}
	}
	
	public void setConnectionURL(String url) {
		closeConnection();
		synchronized (runSignal) {
			runSignal.notifyAll();
		}		
		this.connectionURL = url;
	}
	
	public void closeConnection() {
		try {
			if (iStream != null) iStream.close();
			if (oStream != null) oStream.close();
			if (connection != null) connection.close();
		} catch (IOException e) {
			// Might be closed already, doesn't matter.
		}
		connection = null;
		iStream = null;
		oStream = null;
		synchronized (runSignal) {
			runSignal.notifyAll();
		}		
		status.connected = false;
		notifyStatusUpdate();
	}
	
	void doNextCommand() throws IOException {
		for (Command cmd = getNextCommand(); cmd != null; cmd = getNextCommand()) {
			try {
				cmd.execute(commandTarget);
			} catch (BemusedProtocolException e) {
				throw new IOException("Bemused Exception: "+e.getMessage());
			}
		}
	}
	
	Command getNextCommand() {
		synchronized (cmdQueue) {
			if (cmdQueue.isEmpty() == false) {
				Command ret = (Command)cmdQueue.elementAt(0);
				cmdQueue.removeElementAt(0);
				return ret;
			}
			else return null;
		}
	}
	
	public void destroy() {
    	closeConnection();
    	statusUpdateTimer.cancel();
    	synchronized (runSignal) {
			runSignal.notifyAll();
		}
		try {
			readWriteThread.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
			throw new RuntimeException("InterruptedException caught: "+e.getMessage());	
		}
		statusUpdateListeners.removeAllElements();
	}
	
	public void registerStatusUpdateListener(StatusUpdateListener listener) {
		statusUpdateListeners.addElement(listener);
	}
	
	void notifyStatusUpdate() {
		display.callSerially(new Runnable() { public void run() {
			for (Enumeration e = statusUpdateListeners.elements(); e.hasMoreElements();) {
				StatusUpdateListener listener = (StatusUpdateListener)e.nextElement();
				listener.bemusedStatusChanged(BemusedProtocol.this.status);
			}
		}});
	}
	
	class MyCommandTarget implements CommandTarget {
	
		public DataInputStream getInputStream() {
			return iStream;
		}
	
		public DataOutputStream getOutputStream() {
			return oStream;
		}
		
		public void setPlaylist(String[] items) {
			synchronized (currentStatus) {
				currentStatus.playlist = items;
			}
			notifyStatusUpdate();
		}
		
		public void setDirInfo(byte[] headers, String[] contents) {
			fileBrowser.setDirInfo(headers, contents);
		}
		
		public void setPlaylistPos(int pos) {
			synchronized (currentStatus) {
				currentStatus.playlistPos = pos;
			}
			notifyStatusUpdate();
		}

		public void setInfo(boolean playing, boolean shuffle, boolean repeat, 
				int songLengthSecs, int songPosSecs, String title) {
			synchronized (currentStatus) {
				currentStatus.playing = playing;
				currentStatus.songLengthSecs = songLengthSecs;
				currentStatus.startTime = new Date(new Date().getTime() - songPosSecs*1000);
				String oldTitle = currentStatus.title;
				currentStatus.title = title;
				if (currentStatus.title.compareTo(oldTitle) != 0) {
					updatePlaylistPos();
				}
				currentStatus.shuffle = shuffle;
				currentStatus.repeat = repeat;
			}
			notifyStatusUpdate();
		}
		
		private void updatePlaylistPos() {
			boolean found = true;
			for (int n=0; n<currentStatus.playlist.length; ++n) {
				if (currentStatus.playlist[n].compareTo(currentStatus.title) == 0) {
					currentStatus.playlistPos = n;
					found = true;
				}
			}
			if (found == false) {
				// We couldn't find the current title in the playlist.
				// Maybe it's just written a bit differently, so that's not
				// necessarily an error. 
				// The bemused protocol forces us to get the whole playlist
				// just to get the current playlist position.
				refreshPlaylist();
			}
		}
		
		public void setVolume(int v) {
			if (currentStatus.volume != v) {
				currentStatus.volume = v;
				notifyStatusUpdate();
			}
		}
		
	}

	void appendCommand(Command cmd) {
		synchronized (cmdQueue) {
			cmdQueue.addElement(cmd);
		}
		synchronized (runSignal) {
			runSignal.notifyAll();
		}		
	}

	public void play() {
		appendCommand(new Command("STRT"));
		scheduleStatusUpdate();
	}

	public void stop() {
		appendCommand(new Command("STOP"));
		scheduleStatusUpdate();
	}
	
	public void pause() {
		appendCommand(new Command("PAUS"));
		scheduleStatusUpdate();
	}

	public void next() {
		appendCommand(new Command("NEXT"));
		scheduleStatusUpdate();
	}

	public void previous() {
		appendCommand(new Command("PREV"));
		scheduleStatusUpdate();
	}

	public void shutdownSystem() {
		appendCommand(new Command("SHUT"));
	}

	public void refreshPlaylist() {
		appendCommand(new Command("PLST"));
	}
	
	public void rewind() {
		appendCommand(new Command("RWND"));
		scheduleStatusUpdate();
	}
	
	public void forward() {
		appendCommand(new Command("FFWD"));
		scheduleStatusUpdate();
	}
	
	public void setCurrentTrack(int pos) {
		appendCommand(new Command("SLCT", pos));
		scheduleStatusUpdate();
	}
	
	public void toggleShuffle() {
		byte ns = currentStatus.shuffle ? (byte) 0 : (byte) 1;
		appendCommand(new Command("SHFL", ns));
	}
	
	public void clearPlaylist() {
		appendCommand(new Command("RMAL"));
		refreshPlaylist();
	}
	
	public void toggleRepeat() {
		byte ns = currentStatus.repeat ? (byte) 0 : (byte) 1;
		appendCommand(new Command("REPT", ns));
	}
	
	public void playFile(String filename) {
		int len = filename.length();
		byte[] command = new byte[2];
		command[0] = (byte) (len / 256);
		command[1] = (byte) (len % 256);
		appendCommand(new Command("PLAY", new String(command) + filename));
	}
	
	public String[] playlist() {
		synchronized (currentStatus) {
			return currentStatus.playlist;
		}
	}
	
	public void readDir(String dir) {
		int len = dir.length();
		byte[] command = new byte[2];
		command[0] = (byte) (len / 256);
		command[1] = (byte) (len % 256);
		appendCommand(new Command("DLST", new String(command) + dir));
	}
	
	public void addToPlaylist(String filename) {
		int len = filename.length();
		byte[] command = new byte[2];
		command[0] = (byte) (len / 256);
		command[1] = (byte) (len % 256);
		appendCommand(new Command("LADD", new String(command) + filename));
	}
	
	public PlaybackStatus status() {
		synchronized (currentStatus) {
			return currentStatus;
		}
	}
	
	public int currentTrackNum() {
		return currentStatus.playlistPos;
	}

	/* (non-Javadoc)
	 * @see protocol.MusicPlayer#setVol(boolean, int)
	 */
	public void setVol(boolean relative, int val) {
		int newVal = relative?currentStatus.volume+val:val;
		if (newVal<0) newVal = 0;
		if (newVal>255) newVal = 255;
		currentStatus.volume = newVal;
		scheduleStatusUpdate();
		
		//bbun 28/8/2004 @ 17:10
		// [
			//byte b = newVal>127?(byte)(newVal-255):(byte)newVal; ---> WHY?! it makes the volume wrap again
			byte b = (byte)newVal;
		// ]
		
		appendCommand(new Command("VOLM", b));
	}

}

