/*
 * Created on May 31, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package bemused;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Vector;

public class Command {

	String name;
	String filenameArg;
	int intArg;
	byte byteArg;

	public Command(String name) {
		this.name = name;
	}

	public Command(String name, int intArg) {
		this.name = name;
		this.intArg = intArg;
	}

	public Command(String name, byte byteArg) {
		this.name = name;
		this.byteArg = byteArg;
	}

	public Command(String name, String filenameArg) {
		this.name = name;
		this.filenameArg = filenameArg;
	}

	public void setFilenameArg(String s) {
		filenameArg = s;
	}

	public String getName() {
		return name;
	}

	private byte readByte(DataInputStream input) throws IOException {
		return input.readByte();
	}

	private void readPlaylist(DataInputStream input, CommandTarget target)
			throws IOException {
		Vector playlistVec = new Vector();

		// Read the playlist
		byte[] replyTypeBytes = new byte[7];
		input.readFully(replyTypeBytes);
		// TODO: check if replyTypeBytes contains the sequence "PLSTACK"
		short playlistPos = input.readShort();

		byte[] curItemBuf = new byte[128];
		int curItemBufLen = curItemBuf.length;
		int curItemBufPos = 0;
		byte lastByte;
		do {
			lastByte = input.readByte();
			if (curItemBufPos < curItemBufLen) {
				curItemBuf[curItemBufPos++] = lastByte;
			}
			if (lastByte == 0 || lastByte == (byte) '\n') {
				if (curItemBufPos - 1 > 0) {
					String curItemStr = new String(curItemBuf, 0,
							curItemBufPos - 1, "UTF-8");
					playlistVec.addElement(curItemStr);
				}
				curItemBufPos = 0;
			}
		} while (lastByte != 0);

		int listSize = playlistVec.size();
		String[] ret = new String[listSize < 1 ? 0 : (listSize - 1)];
		// TODO: Find out why there is a dummy item 
		// at the first position of the playlist 

		String playlistItem;
		int retIndex = 0;
		int lastSlash = 0;
		int newLastSlash = 0;
		for (int n = 1; n < listSize; n=n+1) {
			playlistItem = (String) playlistVec.elementAt(n);
			if (playlistItem.startsWith("#")) continue;
			while (newLastSlash > -1)
			{
				lastSlash = newLastSlash;
				newLastSlash = playlistItem.indexOf("\\", lastSlash + 1);
			}
			newLastSlash = 0;
			// dub sez...
			// linuxserver doesnt send slashes; the playlist gets a mess,
			// when only a substring is taken - so use the whole item
			if (lastSlash >0 )
				ret[retIndex] = playlistItem.substring(lastSlash + 1, playlistItem.length()-5);
			else
				ret[retIndex] = playlistItem;
			retIndex++;
		}
		
		String[] retClean = new String[retIndex];
		for (int n = 0; n < retIndex; n++)
		{
			retClean[n] = ret[n];
		}
		
		target.setPlaylist(retClean);
		target.setPlaylistPos(playlistPos);
	}
	
	private void readDir(DataInputStream input, CommandTarget target)
			throws IOException {
		
		byte[] replyTypeBytes = new byte[7];
		input.readFully(replyTypeBytes);
		// TODO: check if replyTypeBytes contains the sequence "LISTACK"
		
		// the first byte now contains info of each entry
		// see: http://bemused.sourceforge.net/book/view/49
		
		byte headerByte;
		byte[] curItemBuf = new byte[128];
		int curItemBufLen = curItemBuf.length;
		int curItemBufPos = 0;
		byte lastByte;
		Vector headerBytes = new Vector();
		Vector dirList = new Vector();

		boolean node_terminator = false;
		
		do {
			headerByte = input.readByte();
			if (((int)headerByte & 240) == 240) node_terminator = true;
			
			do {
				lastByte = input.readByte();
				if (curItemBufPos < curItemBufLen) {
					curItemBuf[curItemBufPos++] = lastByte;
				}
				if (lastByte == 0) {
					if (curItemBufPos - 1 > 0) {
						String curItemStr = new String(curItemBuf, 0, curItemBufPos - 1, "UTF-8");
						headerBytes.addElement(new Byte(headerByte));
						dirList.addElement(curItemStr);
					}
					curItemBufPos = 0;
				}
			} while (lastByte != 0);

		} while (!node_terminator);
		
		int listSize = dirList.size();
		String[] ret_string = new String[listSize < 1 ? 0 : listSize];
		byte[] ret_bytes = new byte[listSize < 1 ? 0 : listSize];
		
		for (int n = 0; n < listSize; n++)
		{
			ret_bytes[n] = ((Byte) headerBytes.elementAt(n)).byteValue();
			ret_string[n] = (String) dirList.elementAt(n);
		}
		target.setDirInfo(ret_bytes, ret_string);
	}
	
	private void readInf2(DataInputStream input, CommandTarget target)
			throws IOException {
		// Read the cmd status name
		byte[] replyTypeBytes = new byte[7];
		//			 TODO: check if replyTypeBytes contains the sequence "INFOACK"
		input.readFully(replyTypeBytes);

		boolean playing = input.readByte() == 0 ? false : true;
		int songLength = input.readInt();
		int songPos = input.readInt();
		boolean shuffle = input.readByte() == 0 ? false : true;
		boolean random = input.readByte() == 0 ? false : true;

		byte nameBuf[] = new byte[64];
		byte lastByte;
		int nameBufPos = 0;
		do {
			lastByte = input.readByte();
			nameBuf[nameBufPos++] = lastByte;
		} while (lastByte != 0 && nameBufPos < nameBuf.length);
		String title = new String(nameBuf, 0, nameBufPos - 1, "UTF-8");

		target.setInfo(playing, shuffle, random, songLength, songPos, title);
	}

	private void readVolume(DataInputStream input, CommandTarget target) 
			throws IOException {
		// Read the cmd status name
		byte[] replyTypeBytes = new byte[7];
		input.readFully(replyTypeBytes);
		byte byteVol = input.readByte();
		int vol = byteVol>=0?byteVol:byteVol+255;
		// Check if the return was GVOLACK or GVOLNAK
		if (new String(replyTypeBytes, 0, 7).compareTo("GVOLACK") == 0) {
			target.setVolume(vol);
		}
	}
	
	private void writeShort(DataOutputStream output, short val)
			throws IOException {
		output.writeShort(val);
	}
	
	private void writeString(DataOutputStream output, String s) throws BemusedProtocolException {
		byte[] stream = s.getBytes();
		try {
			for (int n = 0; n < stream.length; ++n) {
				output.writeByte(stream[n]);
			}
		}
		 catch (IOException e) {
			throw new BemusedProtocolException("BemusedCommand:execute: "
					+ e.getMessage());
		}
	}

	public void execute(CommandTarget target) throws BemusedProtocolException {
		DataOutputStream output = target.getOutputStream();
		DataInputStream input = target.getInputStream();
		try {
			// Send the command name 
			output.write(name.getBytes(), 0, 4);

			if (name.equals("CHCK")) {
				output.flush();
				readByte(input);
			}
			/*else if (name.equals("DINF")) {
			 readDetailedInfoStructure;
			 }*/
			else if (name.equals("DLST")) {
				writeString(output, filenameArg);
				output.flush();
				readDir(input, target);
			 }
			/*else if (name.equals("DOWN")) {
			 sendFilename();
			 readFileStructure();
			 }*/
			else if (name.equals("FADE")) {
				output.flush();
			} else if (name.equals("FFWD")) {
				output.flush();
			}
			/*else if (name.equals("FINF")) {
			 sendFilename();
			 readFileInfoStructure();
			 }*/
			else if (name.equals("GVOL")) {
				output.flush();
				readVolume(input, target);
			}
			/*else if (name.equals("INFO")) {
			 readInfoStructure();
			 }*/
			else if (name.equals("INF2")) {
				output.flush();
				readInf2(input, target);
			}
			else if (name.equals("LADD")) {
			 	writeString(output, filenameArg);
				output.flush();
			 }
			/*else if (name.equals("LIST")) {
				 output.flush();
				 readDirList(input, target);
			 }*/
			else if (name.equals("NEXT")) {
				output.flush();
			} else if (name.equals("PAUS")) {
				output.flush();
			}
			else if (name.equals("PLAY")) {
			 	writeString(output, filenameArg);
				output.flush();
			 }
			/*else if (name.equals("PLEN")) {
			 readPlaylistLength();
			 }*/
			else if (name.equals("PLST")) {
				output.flush();
				readPlaylist(input, target);
			} else if (name.equals("PREV")) {
				output.flush();
			}
			else if (name.equals("REPT")) {
			 	output.writeByte(byteArg);
				output.flush();
			 }
			else if (name.equals("RMAL")) {
				output.flush();
			} else if (name.equals("RWND")) {
				output.flush();
			}
			else if (name.equals("SHFL")) {
			 	output.writeByte(byteArg);
				output.flush();
			 }
			/*else if (name.equals("SEEK")) {
			 writeInt(seekPos);
			 }*/
			else if (name.equals("SHUT")) {
				output.flush();
			} else if (name.equals("STEN")) {
				output.flush();
			} else if (name.equals("SLCT")) {
				writeShort(output, (short) intArg);
				output.flush();
			} else if (name.equals("STOP")) {
				output.flush();
			} else if (name.equals("STRT")) {
				output.flush();
			}
			else if (name.equals("VOLM")) {
				output.writeByte(byteArg);
				output.flush();
			}
			/*else if (name.equals("VERS")) {
			 readVersion();
			 }*/
		} catch (IOException e) {
			throw new BemusedProtocolException("BemusedCommand:execute: "
					+ e.getMessage());
		}
	}

}
